@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU General Public License along
  * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA 02110-1301  USA
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA
  */
 
 #include <config.h>
@@ -44,7 +44,7 @@ get_char (GucharmapCodepointList *list,
   GucharmapBlockCodepointList *block_list = GUCHARMAP_BLOCK_CODEPOINT_LIST (list);
   GucharmapBlockCodepointListPrivate *priv = block_list->priv;
 
-  if (index > (gint)priv->end - priv->start)
+  if (index > (gint) (priv->end - priv->start))
     return (gunichar)(-1);
   else
     return (gunichar) priv->start + index;
