@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU General Public License along
  * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA 02110-1301  USA
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA
  */
 
 #include <config.h>
@@ -21,6 +21,8 @@
 #include <gtk/gtk.h>
 #include <string.h>
 
+#include <glib/gi18n-lib.h>
+
 #include "gucharmap.h"
 #include "gucharmap-private.h"
 
@@ -29,9 +31,7 @@
 #include "unicode-nameslist.h"
 #include "unicode-categories.h"
 #include "unicode-versions.h"
-#if ENABLE_UNIHAN
-# include "unicode-unihan.h"
-#endif
+#include "unicode-unihan.h"
 
 /* constants for hangul (de)composition, see UAX #15 */
 #define SBase 0xAC00
@@ -58,26 +58,37 @@ static const gchar JAMO_T_TABLE[][4] = {
   "S", "SS", "NG", "J", "C", "K", "T", "P", "H"
 };
 
-G_CONST_RETURN gchar *
+const gchar *
 gucharmap_get_unicode_name (gunichar wc)
 {
   static gchar buf[32];
 
   _gucharmap_intl_ensure_initialized ();
 
-  if ((wc >= 0x3400 && wc <= 0x4DB5)
-      || (wc >= 0x4e00 && wc <= 0x9fcb)
-      || (wc >= 0x20000 && wc <= 0x2A6D6)
+  if ((wc >= 0x3400 && wc <= 0x4db5)
+      || (wc >= 0x4e00 && wc <= 0x9fef)
+      || (wc >= 0x20000 && wc <= 0x2a6d6)
       || (wc >= 0x2a700 && wc <= 0x2b734)
-      || (wc >= 0x2b740 && wc <= 0x2b81d))
+      || (wc >= 0x2b740 && wc <= 0x2b81d)
+      || (wc >= 0x2b820 && wc <= 0x2cea1)
+      || (wc >= 0x2ceb0 && wc <= 0x2ebe0))
     {
       g_snprintf (buf, sizeof (buf), "CJK UNIFIED IDEOGRAPH-%04X", wc);
       return buf;
     }
-  /* FIXME: handle U+F900..U+FAFF CJK COMPATIBILITY IDEOGRAPH U+%04X and
-   * U+2F800..U+2FA1D CJK COMPATIBILITY IDEOGRAPH U+%04X here,
-   * instead of via gucharmap_get_unicode_data_name below, to save some space?
-    */
+  else if ((wc >= 0xf900 && wc <= 0xfaff) ||
+           (wc >= 0x2f800 && wc <= 0x2fa1d)) {
+      g_snprintf (buf, sizeof (buf), "CJK COMPATIBILITY IDEOGRAPH-%04X", wc);
+      return buf;
+  }
+  else if (wc >= 0x17000 && wc <= 0x187f1) {
+      g_snprintf (buf, sizeof (buf), "TANGUT IDEOGRAPH-%05X", wc);
+      return buf;
+  }
+  else if (wc >= 0x18800 && wc <= 0x18af2) {
+      g_snprintf (buf, sizeof (buf), "TANGUT COMPONENT-%03u", wc - 0x18800 + 1);
+      return buf;
+  }
   else if (wc >= 0xac00 && wc <= 0xd7af)
     {
       /* compute hangul syllable name as per UAX #15 */
@@ -118,7 +129,7 @@ gucharmap_get_unicode_name (gunichar wc)
     }
 }
 
-G_CONST_RETURN gchar *
+const gchar *
 gucharmap_get_unicode_category_name (gunichar wc)
 {
   _gucharmap_intl_ensure_initialized ();
@@ -160,7 +171,7 @@ gucharmap_get_unicode_category_name (gunichar wc)
 }
 
 /* does a binary search on unicode_names */
-G_CONST_RETURN gchar *
+const gchar *
 gucharmap_get_unicode_data_name (gunichar uc)
 {
   gint min = 0;
@@ -216,7 +227,7 @@ gucharmap_get_unicode_version (gunichar uc)
   return GUCHARMAP_UNICODE_VERSION_UNASSIGNED;
 }
 
-G_CONST_RETURN gchar *
+const gchar *
 gucharmap_unicode_version_to_string (GucharmapUnicodeVersion version)
 {
   g_return_val_if_fail (version >= GUCHARMAP_UNICODE_VERSION_UNASSIGNED &&
@@ -228,8 +239,6 @@ gucharmap_unicode_version_to_string (GucharmapUnicodeVersion version)
   return unicode_version_strings + unicode_version_string_offsets[version - 1];
 }
 
-#if ENABLE_UNIHAN
-
 gint
 gucharmap_get_unihan_count (void)
 {
@@ -238,7 +247,7 @@ gucharmap_get_unihan_count (void)
 
 /* does a binary search; also caches most recent, since it will often be
  * called in succession on the same character */
-static G_CONST_RETURN Unihan *
+static const Unihan *
 _get_unihan (gunichar uc)
 {
   static gunichar most_recent_searched;
@@ -274,61 +283,9 @@ _get_unihan (gunichar uc)
   return NULL;
 }
 
-#else /* #if ENABLE_UNIHAN */
-
-gint
-gucharmap_get_unihan_count ()
-{
-  return 0;
-}
-
-G_CONST_RETURN gchar * 
-gucharmap_get_unicode_kDefinition (gunichar uc)
-{
-  return "This feature was not compiled in.";
-}
-
-G_CONST_RETURN gchar * 
-gucharmap_get_unicode_kCantonese (gunichar uc)
-{
-  return "This feature was not compiled in.";
-}
-
-G_CONST_RETURN gchar * 
-gucharmap_get_unicode_kMandarin (gunichar uc)
-{
-  return "This feature was not compiled in.";
-}
-
-G_CONST_RETURN gchar * 
-gucharmap_get_unicode_kTang (gunichar uc)
-{
-  return "This feature was not compiled in.";
-}
-
-G_CONST_RETURN gchar * 
-gucharmap_get_unicode_kKorean (gunichar uc)
-{
-  return "This feature was not compiled in.";
-}
-
-G_CONST_RETURN gchar * 
-gucharmap_get_unicode_kJapaneseKun (gunichar uc)
-{
-  return "This feature was not compiled in.";
-}
-
-G_CONST_RETURN gchar * 
-gucharmap_get_unicode_kJapaneseOn (gunichar uc)
-{
-  return "This feature was not compiled in.";
-}
-
-#endif /* #else (#if ENABLE_UNIHAN) */
-
 /* does a binary search; also caches most recent, since it will often be
  * called in succession on the same character */
-static G_CONST_RETURN NamesList *
+static const NamesList *
 get_nameslist (gunichar uc)
 {
   static gunichar most_recent_searched;
@@ -401,7 +358,7 @@ gucharmap_get_nameslist_exes (gunichar uc)
  * Returns: (transfer container): newly allocated null-terminated array of gchar*
  * the items are const, but the array should be freed by the caller
  */
-G_CONST_RETURN gchar **
+const gchar **
 gucharmap_get_nameslist_equals (gunichar uc)
 {
   const NamesList *nl;
@@ -432,7 +389,7 @@ gucharmap_get_nameslist_equals (gunichar uc)
  * Returns: (transfer container): newly allocated null-terminated array of gchar*
  * the items are const, but the array should be freed by the caller
  */
-G_CONST_RETURN gchar **
+const gchar **
 gucharmap_get_nameslist_stars (gunichar uc)
 {
   const NamesList *nl;
@@ -463,7 +420,7 @@ gucharmap_get_nameslist_stars (gunichar uc)
  * Returns: (transfer container): newly allocated null-terminated array of gchar*
  * the items are const, but the array should be freed by the caller
  */
-G_CONST_RETURN gchar **
+const gchar **
 gucharmap_get_nameslist_pounds (gunichar uc)
 {
   const NamesList *nl;
@@ -494,7 +451,7 @@ gucharmap_get_nameslist_pounds (gunichar uc)
  * Returns: (transfer container): newly allocated null-terminated array of gchar*
  * the items are const, but the array should be freed by the caller
  */
-G_CONST_RETURN gchar **
+const gchar **
 gucharmap_get_nameslist_colons (gunichar uc)
 {
   const NamesList *nl;
@@ -526,7 +483,7 @@ gucharmap_unichar_validate (gunichar ch)
 }
 
 /**
- * gucharmap_unichar_to_printable_utf8
+ * gucharmap_unichar_to_printable_utf8:
  * @uc: a unicode character 
  * @outbuf: output buffer, must have at least 10 bytes of space.
  *          If %NULL, the length will be computed and returned
@@ -642,8 +599,24 @@ gucharmap_unichar_isgraph (gunichar uc)
 {
   GUnicodeType t = gucharmap_unichar_type (uc);
 
+  /* From http://www.unicode.org/versions/Unicode9.0.0/ch09.pdf, p16
+   * "Unlike most other format control characters, however, they should be
+   *  rendered with a visible glyph, even in circumstances where no suitable
+   *  digit or sequence of digits follows them in logical order."
+   * There the standard talks about the ar signs spanning numbers, but
+   * I think this should apply to all Prepended_Concatenation_Mark format
+   * characters.
+   * Instead of parsing the corresponding data file, just hardcode the
+   * (few!) existing characters here.
+   */
+  if (t == G_UNICODE_FORMAT)
+    return (uc >= 0x0600 && uc <= 0x0605) || 
+	   uc == 0x06DD ||
+           uc == 0x070F ||
+           uc == 0x08E2 ||
+           uc == 0x110BD;
+
   return (t != G_UNICODE_CONTROL
-          && t != G_UNICODE_FORMAT
           && t != G_UNICODE_UNASSIGNED
           && t != G_UNICODE_PRIVATE_USE
           && t != G_UNICODE_SURROGATE
