@@ -1,6 +1,7 @@
 /*
  * Copyright © 2004 Noah Levitt
  * Copyright © 2007, 2008 Christian Persch
+ * Copyright © 2012 Red Hat, Inc.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -14,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License along
  * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA 02110-1301  USA
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA
  */
 
 #include <config.h>
@@ -25,13 +26,6 @@
 #include <glib/gi18n-lib.h>
 #include <gtk/gtk.h>
 
-#if GTK_CHECK_VERSION (2, 90, 7)
-#define GDK_KEY(symbol) GDK_KEY_##symbol
-#else
-#include <gdk/gdkkeysyms.h>
-#define GDK_KEY(symbol) GDK_##symbol
-#endif
-
 #include "gucharmap-print-operation.h"
 #include "gucharmap-search-dialog.h"
 #include "gucharmap-settings.h"
@@ -44,7 +38,7 @@
 static void gucharmap_window_class_init (GucharmapWindowClass *klass);
 static void gucharmap_window_init       (GucharmapWindow *window);
 
-G_DEFINE_TYPE (GucharmapWindow, gucharmap_window, GTK_TYPE_WINDOW)
+G_DEFINE_TYPE (GucharmapWindow, gucharmap_window, GTK_TYPE_APPLICATION_WINDOW)
 
 static void
 show_error_dialog (GtkWindow *parent,
@@ -157,46 +151,23 @@ status_message (GtkWidget       *widget,
     gtk_statusbar_push (GTK_STATUSBAR (guw->status), 0, message);
 }
 
-static gboolean
-update_progress_bar (GucharmapWindow *guw)
-{
-  gdouble fraction_completed;
-
-  fraction_completed = gucharmap_search_dialog_get_completed (GUCHARMAP_SEARCH_DIALOG (guw->search_dialog));
-
-  if (fraction_completed < 0 || fraction_completed > 1)
-    {
-      gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (guw->progress), 0);
-      gtk_progress_bar_set_text (GTK_PROGRESS_BAR (guw->progress), NULL);
-      return FALSE;
-    }
-  else
-    {
-      gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (guw->progress), fraction_completed);
-      return TRUE;
-    }
-}
-
 static void
 search_start (GucharmapSearchDialog *search_dialog,
               GucharmapWindow       *guw)
 {
   GdkCursor *cursor;
-  GtkAction *action;
+  GAction *action;
 
   cursor = gdk_cursor_new_for_display (gtk_widget_get_display (GTK_WIDGET (guw)), GDK_WATCH);
   gdk_window_set_cursor (gtk_widget_get_window (GTK_WIDGET (guw)), cursor);
-  gdk_cursor_unref (cursor);
+  g_object_unref (cursor);
 
-  action = gtk_action_group_get_action (guw->action_group, "Find");
-  gtk_action_set_sensitive (action, FALSE);
-  action = gtk_action_group_get_action (guw->action_group, "FindNext");
-  gtk_action_set_sensitive (action, FALSE);
-  action = gtk_action_group_get_action (guw->action_group, "FindPrevious");
-  gtk_action_set_sensitive (action, FALSE);
-
-  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (guw->progress), _("Searching…"));
-  g_timeout_add (100, (GSourceFunc) update_progress_bar, guw);
+  action = g_action_map_lookup_action (G_ACTION_MAP (guw), "find");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), FALSE);
+  action = g_action_map_lookup_action (G_ACTION_MAP (guw), "find-next");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), FALSE);
+  action = g_action_map_lookup_action (G_ACTION_MAP (guw), "find-previous");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), FALSE);
 }
 
 static void
@@ -204,10 +175,7 @@ search_finish (GucharmapSearchDialog *search_dialog,
                gunichar               found_char,
                GucharmapWindow       *guw)
 {
-  GtkAction *action;
-
-  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (guw->progress), 0);
-  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (guw->progress), NULL);
+  GAction *action;
 
   if (found_char != (gunichar)(-1))
     gucharmap_charmap_set_active_character (guw->charmap, found_char);
@@ -215,48 +183,64 @@ search_finish (GucharmapSearchDialog *search_dialog,
 
   gdk_window_set_cursor (gtk_widget_get_window (GTK_WIDGET (guw)), NULL);
 
-  action = gtk_action_group_get_action (guw->action_group, "Find");
-  gtk_action_set_sensitive (action, TRUE);
-  action = gtk_action_group_get_action (guw->action_group, "FindNext");
-  gtk_action_set_sensitive (action, TRUE);
-  action = gtk_action_group_get_action (guw->action_group, "FindPrevious");
-  gtk_action_set_sensitive (action, TRUE);
+  action = g_action_map_lookup_action (G_ACTION_MAP (guw), "find");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), TRUE);
+  action = g_action_map_lookup_action (G_ACTION_MAP (guw), "find-next");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), TRUE);
+  action = g_action_map_lookup_action (G_ACTION_MAP (guw), "find-previous");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), TRUE);
 }
 
 static void
-search_find (GtkAction       *action, 
-             GucharmapWindow *guw)
+ensure_search_dialog (GucharmapWindow *guw)
 {
-  g_assert (GUCHARMAP_IS_WINDOW (guw));
-
   if (guw->search_dialog == NULL)
     {
       guw->search_dialog = gucharmap_search_dialog_new (guw);
       g_signal_connect (guw->search_dialog, "search-start", G_CALLBACK (search_start), guw);
       g_signal_connect (guw->search_dialog, "search-finish", G_CALLBACK (search_finish), guw);
     }
+}
+
+static void
+search_find (GSimpleAction *action,
+             GVariant      *parameter,
+             gpointer       data)
+{
+  GucharmapWindow *guw = data;
+
+  g_assert (GUCHARMAP_IS_WINDOW (guw));
+
+  ensure_search_dialog (guw);
 
   gucharmap_search_dialog_present (GUCHARMAP_SEARCH_DIALOG (guw->search_dialog));
 }
 
+
 static void
-search_find_next (GtkAction       *action, 
-                  GucharmapWindow *guw)
+search_find_next (GSimpleAction *action,
+                  GVariant      *parameter,
+                  gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   if (guw->search_dialog)
     gucharmap_search_dialog_start_search (GUCHARMAP_SEARCH_DIALOG (guw->search_dialog), GUCHARMAP_DIRECTION_FORWARD);
   else
-    search_find (action, guw);
+    search_find (action, NULL, guw);
 }
 
 static void
-search_find_prev (GtkAction       *action, 
-                  GucharmapWindow *guw)
+search_find_prev (GSimpleAction *action,
+                  GVariant      *parameter,
+                  gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   if (guw->search_dialog)
     gucharmap_search_dialog_start_search (GUCHARMAP_SEARCH_DIALOG (guw->search_dialog), GUCHARMAP_DIRECTION_BACKWARD);
   else
-    search_find (action, guw);
+    search_find (action, NULL, guw);
 }
 
 #ifdef ENABLE_PRINTING
@@ -272,9 +256,12 @@ page_setup_done_cb (GtkPageSetup *page_setup,
 }
 
 static void
-file_page_setup (GtkAction *action,
-                 GucharmapWindow *guw)
+file_page_setup (GSimpleAction *action,
+                 GVariant      *parameter,
+                 gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   ensure_print_data (guw);
 
   gtk_print_run_page_setup_dialog_async (GTK_WINDOW (guw),
@@ -286,68 +273,103 @@ file_page_setup (GtkAction *action,
 
 #if 0
 static void
-file_print_preview (GtkAction *action,
-                    GucharmapWindow *guw)
+file_print_preview (GSimpleAction *action,
+                    GVariant      *parameter,
+                    gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   gucharmap_window_print (guw, GTK_PRINT_OPERATION_ACTION_PREVIEW);
 }
 #endif
 
 static void
-file_print (GtkAction *action,
-            GucharmapWindow *guw)
+file_print (GSimpleAction *action,
+            GVariant      *parameter,
+            gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   gucharmap_window_print (guw, GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG);
 }
 
 #endif /* ENABLE_PRINTING */
 
 static void
-close_window (GtkAction *action,
-              GtkWidget *widget)
+close_window (GSimpleAction *action,
+              GVariant      *parameter,
+              gpointer       data)
 {
+  GtkWidget *widget = data;
+
   gtk_widget_destroy (widget);
 }
 
 static void
-font_bigger (GtkAction       *action, 
-             GucharmapWindow *guw)
+font_bigger (GSimpleAction *action,
+             GVariant      *parameter,
+             gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   gucharmap_mini_font_selection_change_font_size (GUCHARMAP_MINI_FONT_SELECTION (guw->fontsel),
                                                   FONT_CHANGE_FACTOR);
 }
 
 static void
-font_smaller (GtkAction       *action, 
-              GucharmapWindow *guw)
+font_smaller (GSimpleAction *action,
+              GVariant      *parameter,
+              gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   gucharmap_mini_font_selection_change_font_size (GUCHARMAP_MINI_FONT_SELECTION (guw->fontsel),
                                                   1.0f / FONT_CHANGE_FACTOR);
 }
 
 static void
-font_default (GtkAction       *action, 
-              GucharmapWindow *guw)
+font_default (GSimpleAction *action,
+              GVariant      *parameter,
+              gpointer       data)
 {
+  GucharmapWindow *guw = data;
+
   gucharmap_mini_font_selection_reset_font_size (GUCHARMAP_MINI_FONT_SELECTION (guw->fontsel));
 }
 
 static void
-snap_cols_pow2 (GtkAction        *action, 
-                GucharmapWindow  *guw)
+snap_cols_pow2_changed (GSettings  *settings,
+                        const char *key,
+                        gpointer    data)
 {
-  gboolean is_active = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
-  gucharmap_charmap_set_snap_pow2 (guw->charmap, is_active);
-  gucharmap_settings_set_snap_pow2 (is_active);
+  GucharmapWindow  *guw = data;
+
+  gucharmap_charmap_set_snap_pow2 (guw->charmap,
+                                   g_settings_get_boolean (settings, key));
 }
 
 static void
-no_font_fallback_toggled_cb (GtkAction       *action, 
-                             GucharmapWindow *guw)
+toggle_action_activated (GSimpleAction *action,
+                         GVariant      *parameter,
+                         gpointer       data)
 {
-  gboolean is_active = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
+  GVariant *state = g_action_get_state (G_ACTION (action));
+  gboolean value = g_variant_get_boolean (state);
+
+  g_action_change_state (G_ACTION (action), g_variant_new_boolean (!value));
+  g_variant_unref (state);
+}
+
+static void
+change_no_font_fallback (GSimpleAction *action,
+                         GVariant      *state,
+                         gpointer       data)
+{
+  GucharmapWindow *guw = data;
+  gboolean is_active = g_variant_get_boolean (state);
 
   gucharmap_charmap_set_font_fallback (guw->charmap, !is_active);
+  g_simple_action_set_state (action, state);
 /*  gucharmap_settings_set_font_fallback (is_active); */
 }
 
@@ -371,78 +393,29 @@ open_url (GtkWindow *parent,
 }
 
 static void
-help_contents (GtkAction *action,
-               GucharmapWindow *window)
+help_contents (GSimpleAction *action,
+               GVariant      *parameter,
+               gpointer       data)
 {
-  const char *lang;
-  char *uri = NULL, *url;
-  guint i;
- 
-  const char * const * langs = g_get_language_names ();
-  for (i = 0; langs[i]; i++) {
-    lang = langs[i];
-    if (strchr (lang, '.')) {
-      continue;
-    }
- 
-    uri = g_build_filename (HELPDIR,
-                            "gucharmap", /* DOC_MODULE */
-                            lang,
-                            "gucharmap.xml",
-                            NULL);
-					
-    if (g_file_test (uri, G_FILE_TEST_EXISTS)) {
-      break;
-    }
-
-    g_free (uri);
-    uri = NULL;
-  }
-
-  if (!uri)
-    return;
-
-  url = g_strconcat ("ghelp://", uri, NULL);
-  open_url (GTK_WINDOW (window), url, gtk_get_current_event_time ());
-  g_free (uri);
-  g_free (url);
-}
-
-#if !GTK_CHECK_VERSION (2, 90, 8)
-static void
-about_open_url (GtkAboutDialog *about,
-                const char *link,
-                gpointer data)
-{
-  open_url (GTK_WINDOW (about), link, gtk_get_current_event_time ());
+  GucharmapWindow *window = data;
+  open_url (GTK_WINDOW (window),
+            "help:gucharmap", /* DOC_MODULE */
+            gtk_get_current_event_time ());
 }
 
 static void
-about_email_hook (GtkAboutDialog *about,
-		  const char *email_address,
-		  gpointer user_data)
-{
-  char *escaped, *uri;
-
-  escaped = g_uri_escape_string (email_address, NULL, FALSE);
-  uri = g_strdup_printf ("mailto:%s", escaped);
-  g_free (escaped);
-
-  open_url (GTK_WINDOW (about), uri, gtk_get_current_event_time ());
-  g_free (uri);
-}
-#endif
-
-static void
-help_about (GtkAction       *action, 
-            GucharmapWindow *guw)
+help_about (GSimpleAction *action,
+            GVariant      *parameter,
+            gpointer       data)
 {
+  GucharmapWindow *guw = data;
   const gchar *authors[] =
     { 
       "Noah Levitt <nlevitt@columbia.edu>", 
       "Daniel Elstner <daniel.elstner@gmx.net>", 
       "Padraig O'Briain <Padraig.Obriain@sun.com>",
-      "Christian Persch <" "chpe" "\100" "gnome" "." "org" ">",
+      "Christian Persch <" "chpe" "\100" "src.gnome" "." "org" ">",
+      "DaeHyun Sung <sungdh86@gmail.com>",
       NULL 
     };
 
@@ -468,81 +441,79 @@ help_about (GtkAction       *action,
        "the GNU General Public License and Unicode Copyright for more details."),
     N_("You should have received a copy of the GNU General Public License "
        "along with Gucharmap; if not, write to the Free Software Foundation, Inc., "
-       "59 Temple Place, Suite 330, Boston, MA  02110-1301  USA"),
+       "51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA"),
     N_("Also you should have received a copy of the Unicode Copyright along "
        "with Gucharmap; you can always find it at Unicode's website: "
-       "http://www.unicode.org/copyright.html")
+       "https://www.unicode.org/copyright.html")
   };
   gchar *license_trans;
   license_trans = g_strconcat (_(license[0]), "\n\n", _(license[1]), "\n\n",
 			       _(license[2]), "\n\n", _(license[3]), "\n\n",
 			       _(license[4]), "\n\n", NULL);
 
-#if !GTK_CHECK_VERSION (2, 90, 8)
-  gtk_about_dialog_set_url_hook (about_open_url, NULL, NULL);
-  gtk_about_dialog_set_email_hook (about_email_hook, NULL, NULL);
-#endif
-
   gtk_show_about_dialog (GTK_WINDOW (guw),
 			 "program-name", _("GNOME Character Map"),
 			 "version", VERSION,
-			 "comments", _("Based on the Unicode Character Database 5.2"),
+			 "comments", _("Based on the Unicode Character Database 11.0.0"),
 			 "copyright", "Copyright © 2004 Noah Levitt\n"
-				      "Copyright © 1991–2009 Unicode, Inc.\n"
-				      "Copyright © 2007–2010 Christian Persch",
+				      "Copyright © 1991–2018 Unicode, Inc.\n"
+				      "Copyright © 2007–2018 Christian Persch\n"
+                                      "Copyright © 2016 DaeHyun Sung",
 			 "documenters", documenters,
 			 "license", license_trans,
 			 "wrap-license", TRUE,
 			 "logo-icon-name", GUCHARMAP_ICON_NAME,
   			 "authors", authors,
 			 "translator-credits", _("translator-credits"),
-			 "website", "http://live.gnome.org/Gucharmap",
+			 "website", "https://wiki.gnome.org/Apps/Gucharmap",
 			 NULL);
 
   g_free (license_trans);
 }
 
 static void
-next_or_prev_character (GtkAction       *action,
-                        GucharmapWindow *guw)
+next_or_prev_character (GSimpleAction *action,
+                        GVariant      *parameter,
+                        gpointer    data)
 {
+  GucharmapWindow *guw = data;
   GucharmapChartable *chartable;
   GucharmapChartableClass *klass;
-  GtkBindingSet *binding_set;
   const char *name;
   guint keyval = 0;
 
-  name = gtk_action_get_name (action);
-  if (strcmp (name, "NextCharacter") == 0) {
-    keyval = GDK_KEY (Right);
-  } else if (strcmp (name, "PreviousCharacter") == 0) {
-    keyval = GDK_KEY (Left);
+  name = g_action_get_name (G_ACTION (action));
+  if (strcmp (name, "next-character") == 0) {
+    keyval = GDK_KEY_Right;
+  } else if (strcmp (name, "previous-character") == 0) {
+    keyval = GDK_KEY_Left;
+  } else {
+    g_assert_not_reached ();
   }
 
   chartable = gucharmap_charmap_get_chartable (guw->charmap);
   klass = GUCHARMAP_CHARTABLE_GET_CLASS (chartable);
-  binding_set = gtk_binding_set_by_class (klass);
   gtk_binding_set_activate (gtk_binding_set_by_class (klass),
                             keyval,
                             0,
-#if GTK_CHECK_VERSION (2, 91, 0)
                             G_OBJECT (chartable));
-#else
-                            GTK_OBJECT (chartable));
-#endif
 }
 
 static void
-next_chapter (GtkAction       *action,
-              GucharmapWindow *guw)
+next_chapter (GSimpleAction *action,
+              GVariant      *parameter,
+              gpointer    data)
 {
+  GucharmapWindow *guw = data;
   gucharmap_charmap_next_chapter (guw->charmap);
 }
 
 static void
-prev_chapter (GtkAction       *action,
-              GucharmapWindow *guw)
+prev_chapter (GSimpleAction *action,
+              GVariant      *parameter,
+              gpointer    data)
 {
+  GucharmapWindow *guw = data;
   gucharmap_charmap_previous_chapter (guw->charmap);
 }
 
@@ -551,12 +522,20 @@ chapters_set_labels (const gchar     *labelnext,
 		     const gchar     *labelprev,
 		     GucharmapWindow *guw)
 {
-  GtkAction *action;
+  GtkApplication *app;
+  GMenuModel *model;
+  int n_items;
 
-  action = gtk_action_group_get_action (guw->action_group, "NextChapter");
-  g_object_set ( G_OBJECT (action), "label", labelnext, NULL);
-  action = gtk_action_group_get_action (guw->action_group, "PreviousChapter");
-  g_object_set ( G_OBJECT (action), "label", labelprev, NULL);
+  app = GTK_APPLICATION (g_application_get_default ());
+  g_return_if_fail (app != NULL);
+  model = G_MENU_MODEL (g_object_get_data (G_OBJECT (app), "go-chapter-menu"));
+
+  n_items = g_menu_model_get_n_items (model);
+  while (n_items--)
+    g_menu_remove (G_MENU (model), 0);
+
+  g_menu_append (G_MENU (model), labelnext, "win.next-chapter");
+  g_menu_append (G_MENU (model), labelprev, "win.previous-chapter");
 }
 
 enum {
@@ -591,35 +570,22 @@ gucharmap_window_set_chapters_model (GucharmapWindow *guw,
 }
 
 static void
-view_by (GtkAction        *action,
-	 GtkRadioAction   *radioaction,
-         GucharmapWindow  *guw)
+gucharmap_window_group_by_changed (GSettings   *settings,
+                                   const gchar *key,
+                                   gpointer     user_data)
 {
-  GucharmapChaptersMode mode;
+  GucharmapWindow *guw = user_data;
 
-  switch (gtk_radio_action_get_current_value (radioaction))
-    {
-      case VIEW_BY_SCRIPT:
-        mode = GUCHARMAP_CHAPTERS_SCRIPT;
-	break;
-      
-      case VIEW_BY_BLOCK:
-        mode = GUCHARMAP_CHAPTERS_BLOCK;
-	break;
-      
-      default:
-        g_assert_not_reached ();
-    }
-
-  gucharmap_window_set_chapters_model (guw, mode);
-  gucharmap_settings_set_chapters_mode (mode);
+  gucharmap_window_set_chapters_model (guw, g_settings_get_enum (settings, "group-by"));
 }
 
 #ifdef DEBUG_chpe
 static void
-move_to_next_screen_cb (GtkAction *action,
-                        GtkWidget *widget)
+move_to_next_screen_cb (GSimpleAction *action,
+                        GVariant      *parameter,
+                        gpointer       data)
 {
+  GtkWidget *widget = data;
   GdkScreen *screen;
   GdkDisplay *display;
   int number_of_screens, screen_num;
@@ -639,53 +605,6 @@ move_to_next_screen_cb (GtkAction *action,
 }
 #endif
 
-/* create the menu entries */
-
-static const char ui_info [] =
-  "<menubar name='MenuBar'>"
-    "<menu action='File'>"
-#ifdef ENABLE_PRINTING
-      "<menuitem action='PageSetup' />"
-#if 0
-      "<menuitem action='PrintPreview' />"
-#endif
-      "<menuitem action='Print' />"
-      "<separator />"
-#endif /* ENABLE_PRINTING */
-      "<menuitem action='Close' />"
-#ifdef DEBUG_chpe
-      "<menuitem action='MoveNextScreen' />"
-#endif
-    "</menu>"
-    "<menu action='View'>"
-      "<menuitem action='ByScript' />"
-      "<menuitem action='ByUnicodeBlock' />"
-      "<separator />"
-      "<menuitem action='ShowOnlyGlyphsInFont' />"
-      "<menuitem action='SnapColumns' />"
-      "<separator />"
-      "<menuitem action='ZoomIn' />"
-      "<menuitem action='ZoomOut' />"
-      "<menuitem action='NormalSize' />"
-    "</menu>"
-    "<menu action='Search'>"
-      "<menuitem action='Find' />"
-      "<menuitem action='FindNext' />"
-      "<menuitem action='FindPrevious' />"
-    "</menu>"
-    "<menu action='Go'>"
-      "<menuitem action='NextCharacter' />"
-      "<menuitem action='PreviousCharacter' />"
-      "<separator />"
-      "<menuitem action='NextChapter' />"
-      "<menuitem action='PreviousChapter' />"
-    "</menu>"
-    "<menu action='Help'>"
-      "<menuitem action='HelpContents' />"
-      "<menuitem action='About' />"
-    "</menu>"
-  "</menubar>";
-
 static void
 insert_character_in_text_to_copy (GucharmapChartable *chartable,
                                   GucharmapWindow *guw)
@@ -729,33 +648,16 @@ status_realize (GtkWidget       *status,
                 GucharmapWindow *guw)
 {
   GtkAllocation *allocation;
-#if GTK_CHECK_VERSION (2, 18, 0)
   GtkAllocation widget_allocation;
 
   gtk_widget_get_allocation (guw->status, &widget_allocation);
   allocation = &widget_allocation;
-#else
-  allocation = &guw->status->allocation;
-#endif
 
   /* FIXMEchpe ewww... */
   /* increase the height a bit so it doesn't resize itself */
   gtk_widget_set_size_request (guw->status, -1, allocation->height + 9);
 }
 
-static gboolean
-save_last_char_idle_cb (GucharmapWindow *guw)
-{
-  gunichar wc;
-
-  guw->save_last_char_idle_id = 0;
-
-  wc = gucharmap_charmap_get_active_character (guw->charmap);
-  gucharmap_settings_set_last_char (wc);
-
-  return FALSE;
-}
-
 static void
 fontsel_sync_font_desc (GucharmapMiniFontSelection *fontsel,
                         GParamSpec *pspec,
@@ -774,7 +676,7 @@ fontsel_sync_font_desc (GucharmapMiniFontSelection *fontsel,
   guw->in_notification = FALSE;
 
   font = pango_font_description_to_string (font_desc);
-  gucharmap_settings_set_font (font);
+  g_settings_set (guw->settings, "font", "ms", font);
   g_free (font);
 }
 
@@ -796,224 +698,177 @@ charmap_sync_font_desc (GucharmapCharmap *charmap,
   guw->in_notification = FALSE;
 }
 
+static int last_char_dirty = 0;
+
 static void
-charmap_sync_active_character (GtkWidget *widget,
-                               GParamSpec *pspec,
+charmap_dirty_active_character (GtkWidget *widget,
+                                GParamSpec *pspec,
+                                GucharmapWindow *guw)
+{
+  last_char_dirty = 1;
+}
+
+static gboolean
+charmap_save_active_character (GtkWidget *widget,
                                GucharmapWindow *guw)
 {
-  if (guw->save_last_char_idle_id != 0)
-    return;
+  if (last_char_dirty)
+    g_settings_set_uint (guw->settings, "last-char", 
+                         gucharmap_charmap_get_active_character (guw->charmap));
 
-  guw->save_last_char_idle_id = g_idle_add ((GSourceFunc) save_last_char_idle_cb, guw);
+  return FALSE;
 }
 
 static void
 gucharmap_window_init (GucharmapWindow *guw)
 {
-  GtkWidget *big_vbox, *hbox, *bbox, *button, *label;
+  GtkWidget *grid, *button, *label;
   GucharmapChartable *chartable;
   /* tooltips are NULL because they are never actually shown in the program */
-  const GtkActionEntry menu_entries[] =
+  const GActionEntry menu_entries[] =
   {
-    { "File", NULL, N_("_File"), NULL, NULL, NULL },
-    { "View", NULL, N_("_View"), NULL, NULL, NULL },
-    { "Search", NULL, N_("_Search"), NULL, NULL, NULL },
-    { "Go", NULL, N_("_Go"), NULL, NULL, NULL },
-    { "Help", NULL, N_("_Help"), NULL, NULL, NULL },
-
 #ifdef ENABLE_PRINTING
-    { "PageSetup", NULL, N_("Page _Setup"), NULL,
-      NULL, G_CALLBACK (file_page_setup) },
+    { "page-setup", file_page_setup, NULL, NULL, NULL },
 #if 0
-    { "PrintPreview", GTK_STOCK_PRINT_PREVIEW, NULL, NULL,
-      NULL, G_CALLBACK (file_print_preview) },
+    { "print-preview", file_print_preview, NULL, NULL, NULL },
 #endif
-    { "Print", GTK_STOCK_PRINT, NULL, NULL,
-      NULL, G_CALLBACK (file_print) },
+    { "print", file_print, NULL, NULL, NULL },
 #endif /* ENABLE_PRINTING */
-    { "Close", GTK_STOCK_CLOSE, NULL, NULL,
-      NULL, G_CALLBACK (close_window) },
+    { "close", close_window, NULL, NULL, NULL },
 
-    { "ZoomIn", GTK_STOCK_ZOOM_IN, NULL, "<control>plus",
-      NULL, G_CALLBACK (font_bigger) },
-    { "ZoomOut", GTK_STOCK_ZOOM_OUT, NULL, "<control>minus",
-      NULL, G_CALLBACK (font_smaller) },
-    { "NormalSize", GTK_STOCK_ZOOM_100, NULL, "<control>0",
-      NULL, G_CALLBACK (font_default) },
+    { "zoom-in", font_bigger, NULL, NULL, NULL },
+    { "zoom-out", font_smaller, NULL, NULL, NULL },
+    { "normal-size", font_default, NULL, NULL, NULL },
 
-    { "Find", GTK_STOCK_FIND, NULL, NULL,
-      NULL, G_CALLBACK (search_find) },
-    { "FindNext", NULL, N_("Find _Next"), "<control>G",
-      NULL, G_CALLBACK (search_find_next) },
-    { "FindPrevious", NULL, N_("Find _Previous"), "<shift><control>G",
-      NULL, G_CALLBACK (search_find_prev) },
+    { "find", search_find, NULL, NULL, NULL },
+    { "find-next", search_find_next, NULL, NULL, NULL },
+    { "find-previous", search_find_prev, NULL, NULL, NULL },
 
-    { "NextCharacter", NULL, N_("_Next Character"), "<control>N",
-      NULL, G_CALLBACK (next_or_prev_character) },
-    { "PreviousCharacter", NULL, N_("_Previous Character"), "<control>P",
-      NULL, G_CALLBACK (next_or_prev_character) },
-    { "NextChapter", NULL, N_("Next Script"), "<control>Page_Down",
-      NULL, G_CALLBACK (next_chapter) },
-    { "PreviousChapter", NULL, N_("Previous Script"), "<control>Page_Up",
-      NULL, G_CALLBACK (prev_chapter) },
+    { "next-character", next_or_prev_character, NULL, NULL, NULL },
+    { "previous-character", next_or_prev_character, NULL, NULL, NULL },
+    { "next-chapter", next_chapter, NULL, NULL, NULL },
+    { "previous-chapter", prev_chapter, NULL, NULL, NULL },
 
-    { "HelpContents", GTK_STOCK_HELP, N_("_Contents"), "F1",
-      NULL, G_CALLBACK (help_contents) },
-    { "About", GTK_STOCK_ABOUT, N_("_About"), NULL,
-      NULL, G_CALLBACK (help_about) },
+    { "help", help_contents, NULL, NULL, NULL },
+    { "about", help_about, NULL, NULL, NULL },
 
   #ifdef DEBUG_chpe
-    { "MoveNextScreen", NULL, "Move window to next screen", NULL,
-      NULL, G_CALLBACK (move_to_next_screen_cb) },
+    { "move-next-screen", move_to_next_screen_cb, NULL, NULL, NULL },
   #endif
+
+    { "show-only-glyphs-in-font", toggle_action_activated, NULL, "false",
+      change_no_font_fallback },
   };
-  const GtkRadioActionEntry radio_menu_entries[] =
-  {
-    { "ByScript", NULL, N_("By _Script"), NULL,
-      NULL, VIEW_BY_SCRIPT },
-    { "ByUnicodeBlock", NULL, N_("By _Unicode Block"), NULL,
-      NULL, VIEW_BY_BLOCK }
-  };
-  const GtkToggleActionEntry toggle_menu_entries[] =
-  {
-    { "ShowOnlyGlyphsInFont", NULL, N_("Sho_w only glyphs from this font"), NULL,
-      NULL,
-      G_CALLBACK (no_font_fallback_toggled_cb), FALSE },
-    { "SnapColumns", NULL, N_("Snap _Columns to Power of Two"), NULL,
-      NULL,
-      G_CALLBACK (snap_cols_pow2), FALSE },
-  };
-  GtkWidget *menubar;
-  GtkAction *action;
+  GAction *action;
+  gunichar active;
+  gchar *font;
+
+  guw->settings = g_settings_new ("org.gnome.Charmap");
 
   gtk_window_set_title (GTK_WINDOW (guw), _("Character Map"));
   gtk_window_set_icon_name (GTK_WINDOW (guw), GUCHARMAP_ICON_NAME);
 
-  /* UI manager setup */
-  guw->uimanager = gtk_ui_manager_new();
+  g_action_map_add_action_entries (G_ACTION_MAP (guw),
+                                   menu_entries, G_N_ELEMENTS (menu_entries),
+                                   guw);
 
-  gtk_window_add_accel_group ( GTK_WINDOW (guw),
-  			       gtk_ui_manager_get_accel_group (guw->uimanager) );
-  
-  guw->action_group = gtk_action_group_new ("gucharmap_actions");
-  gtk_action_group_set_translation_domain (guw->action_group, GETTEXT_PACKAGE);
-
-  gtk_action_group_add_actions (guw->action_group,
-  				menu_entries,
-				G_N_ELEMENTS (menu_entries),
-				guw);
-  gtk_action_group_add_radio_actions (guw->action_group,
-  				      radio_menu_entries,
-				      G_N_ELEMENTS (radio_menu_entries),
-				      gucharmap_settings_get_chapters_mode(),
-				      G_CALLBACK (view_by),
-				      guw);
-  gtk_action_group_add_toggle_actions (guw->action_group,
-  				       toggle_menu_entries,
-				       G_N_ELEMENTS (toggle_menu_entries),
-				       guw);
-
-  gtk_ui_manager_insert_action_group (guw->uimanager, guw->action_group, 0);
-  g_object_unref (guw->action_group);
-  
-  gtk_ui_manager_add_ui_from_string (guw->uimanager, ui_info, strlen (ui_info), NULL);
-  
   /* Now the widgets */
-  big_vbox = gtk_vbox_new (FALSE, 0);
-  gtk_container_add (GTK_CONTAINER (guw), big_vbox);
-
-  /* First the menubar */
-  menubar = gtk_ui_manager_get_widget (guw->uimanager, "/MenuBar");
-  gtk_box_pack_start (GTK_BOX (big_vbox), menubar, FALSE, FALSE, 0);
+  grid = gtk_grid_new ();
+  gtk_container_add (GTK_CONTAINER (guw), grid);
 
   /* The font selector */
   guw->fontsel = gucharmap_mini_font_selection_new ();
-  g_signal_connect (guw->fontsel, "notify::font-desc",
-                    G_CALLBACK (fontsel_sync_font_desc), guw);
-  gtk_box_pack_start (GTK_BOX (big_vbox), guw->fontsel, FALSE, FALSE, 0);
+  gtk_grid_attach (GTK_GRID (grid), guw->fontsel, 0, 1, 3, 1);
   gtk_widget_show (GTK_WIDGET (guw->fontsel));
 
   /* The charmap */
   guw->charmap = GUCHARMAP_CHARMAP (gucharmap_charmap_new ());
-  g_signal_connect (guw->charmap, "notify::active-character",
-                    G_CALLBACK (charmap_sync_active_character), guw);
   g_signal_connect (guw->charmap, "notify::font-desc",
                     G_CALLBACK (charmap_sync_font_desc), guw);
 
-  gtk_box_pack_start (GTK_BOX (big_vbox), GTK_WIDGET (guw->charmap),
-                      TRUE, TRUE, 0);
+  gtk_grid_attach (GTK_GRID (grid), GTK_WIDGET (guw->charmap), 0, 2, 3, 1);
   gtk_widget_show (GTK_WIDGET (guw->charmap));
 
   /* Text to copy entry + button */
-  hbox = gtk_hbox_new (FALSE, 6);
-  gtk_container_set_border_width (GTK_CONTAINER (hbox), 6);
-
   label = gtk_label_new_with_mnemonic (_("_Text to copy:"));
-  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
+  g_object_set (label, "margin", 6, NULL);
+  gtk_grid_attach (GTK_GRID (grid), label, 0, 3, 1, 1);
   gtk_widget_show (label);
 
-  bbox = gtk_hbutton_box_new ();
-
   button = gtk_button_new_from_stock (GTK_STOCK_COPY);
+  g_object_set (button, "margin", 6, NULL);
   gtk_widget_set_tooltip_text (button, _("Copy to the clipboard."));
   g_signal_connect (G_OBJECT (button), "clicked",
                     G_CALLBACK (edit_copy), guw);
-  gtk_box_pack_start (GTK_BOX (bbox), button, FALSE, FALSE, 0);
+  gtk_grid_attach (GTK_GRID (grid), button, 2, 3, 1, 1);
   gtk_widget_show (button);
 
   gtk_widget_set_sensitive (button, FALSE);
   guw->text_to_copy_entry = gtk_entry_new ();
+  g_object_set (guw->text_to_copy_entry, "margin", 6, NULL);
+  gtk_widget_set_hexpand (guw->text_to_copy_entry, TRUE);
   gtk_label_set_mnemonic_widget (GTK_LABEL (label), guw->text_to_copy_entry);
   g_signal_connect (G_OBJECT (guw->text_to_copy_entry), "changed",
                     G_CALLBACK (entry_changed_sensitize_button), button);
 
-  gtk_box_pack_start (GTK_BOX (hbox), guw->text_to_copy_entry, TRUE, TRUE, 0);
+  gtk_grid_attach (GTK_GRID (grid), guw->text_to_copy_entry, 1, 3, 1, 1);
   gtk_widget_show (guw->text_to_copy_entry);
 
-  gtk_box_pack_end (GTK_BOX (hbox), bbox, FALSE, FALSE, 0);
-  gtk_widget_show (bbox);
-
-  gtk_box_pack_start (GTK_BOX (big_vbox), hbox, FALSE, FALSE, 0);
-  gtk_widget_show (hbox);
-
   /* FIXMEchpe!! */
   chartable = gucharmap_charmap_get_chartable (guw->charmap);
   g_signal_connect (chartable, "activate", G_CALLBACK (insert_character_in_text_to_copy), guw);
 
   /* Finally the statusbar */
-  hbox = gtk_hbox_new (FALSE, 0);
-  gtk_box_pack_start (GTK_BOX (big_vbox), hbox, FALSE, FALSE, 0);
-
   guw->status = gtk_statusbar_new ();
-  gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (guw->status), FALSE);
-  gtk_box_pack_start (GTK_BOX (hbox), guw->status, TRUE, TRUE, 0);
+  gtk_grid_attach (GTK_GRID (grid), guw->status, 0, 4, 3, 1);
   gtk_widget_show (guw->status);
   g_signal_connect (guw->status, "realize", G_CALLBACK (status_realize), guw);
 
-  guw->progress = gtk_progress_bar_new ();
-  gtk_box_pack_start (GTK_BOX (hbox), guw->progress, FALSE, FALSE, 0);
-
-#if 0
-  grip = gtk_statusbar_new ();
-  gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (grip), TRUE);
-  gtk_box_pack_start (GTK_BOX (hbox), grip, FALSE, FALSE, 0);
-#endif
-  gtk_widget_show_all (hbox);
-
   g_signal_connect (guw->charmap, "status-message", G_CALLBACK (status_message), guw);
 
-  gtk_widget_show (big_vbox);
-
-  action = gtk_action_group_get_action (guw->action_group, "SnapColumns");
-  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
-                                gucharmap_settings_get_snap_pow2 ());
-
-  gucharmap_window_set_chapters_model (guw, gucharmap_settings_get_chapters_mode ());
-
-  gucharmap_charmap_set_active_character (guw->charmap,
-                                          gucharmap_settings_get_last_char ());
+  gtk_widget_show (grid);
 
   gtk_widget_grab_focus (GTK_WIDGET (gucharmap_charmap_get_chartable (guw->charmap)));
+
+  gtk_window_set_has_resize_grip (GTK_WINDOW (guw), TRUE);
+
+  /* snap-to-power-of-two */
+  action = g_settings_create_action (guw->settings, "snap-cols-pow2");
+  g_action_map_add_action (G_ACTION_MAP (guw), action);
+  snap_cols_pow2_changed (guw->settings, "snap-cols-pow2", guw);
+  g_signal_connect (guw->settings, "changed::snap-cols-pow2",
+                    G_CALLBACK (snap_cols_pow2_changed), guw);
+
+  /* read initial settings */
+  /* font */
+  g_settings_get (guw->settings, "font", "ms", &font);
+  if (font != NULL)
+    {
+      gucharmap_window_set_font (guw, font);
+      g_free (font);
+    }
+
+  /* group by */
+  g_action_map_add_action (G_ACTION_MAP (guw), g_settings_create_action (guw->settings, "group-by"));
+  g_signal_connect_object (guw->settings, "changed::group-by", G_CALLBACK (gucharmap_window_group_by_changed), guw, 0);
+  gucharmap_window_set_chapters_model (guw, g_settings_get_enum (guw->settings, "group-by"));
+
+  /* active character */
+  active = g_settings_get_uint (guw->settings, "last-char");
+  gucharmap_charmap_set_active_character (guw->charmap, active);
+
+  /* window geometry */
+  gucharmap_settings_add_window (GTK_WINDOW (guw));
+
+  /* connect these only after applying the initial settings in order to
+   * avoid unnecessary writes to GSettings.
+   */
+  g_signal_connect (guw, "destroy",
+                    G_CALLBACK (charmap_save_active_character), guw);
+  g_signal_connect (guw->charmap, "notify::active-character",
+                    G_CALLBACK (charmap_dirty_active_character), guw);
+  g_signal_connect (guw->fontsel, "notify::font-desc",
+                    G_CALLBACK (fontsel_sync_font_desc), guw);
 }
 
 static void
@@ -1021,9 +876,6 @@ gucharmap_window_finalize (GObject *object)
 {
   GucharmapWindow *guw = GUCHARMAP_WINDOW (object);
 
-  if (guw->save_last_char_idle_id != 0)
-    g_source_remove (guw->save_last_char_idle_id);
-
   if (guw->page_setup)
     g_object_unref (guw->page_setup);
 
@@ -1044,9 +896,10 @@ gucharmap_window_class_init (GucharmapWindowClass *klass)
 /* Public API */
 
 GtkWidget *
-gucharmap_window_new (void)
+gucharmap_window_new (GtkApplication *application)
 {
-  return GTK_WIDGET (g_object_new (gucharmap_window_get_type (), NULL));
+  return GTK_WIDGET (g_object_new (gucharmap_window_get_type (),
+                     "application", application, NULL));
 }
 
 void
@@ -1057,11 +910,7 @@ gucharmap_window_set_font (GucharmapWindow *guw,
 
   g_return_if_fail (GUCHARMAP_IS_WINDOW (guw));
 
-#if GTK_CHECK_VERSION (2,20,0)
   g_assert (!gtk_widget_get_realized (GTK_WIDGET (guw)));
-#else
-  g_assert (!GTK_WIDGET_REALIZED (guw));
-#endif
 
   if (!font)
     return;
@@ -1070,3 +919,16 @@ gucharmap_window_set_font (GucharmapWindow *guw,
   gucharmap_charmap_set_font_desc (guw->charmap, font_desc);
   pango_font_description_free (font_desc);
 }
+
+void
+gucharmap_window_search (GucharmapWindow *guw,
+                         const char *str)
+{
+  g_return_if_fail (GUCHARMAP_IS_WINDOW (guw));
+  g_return_if_fail (str != NULL);
+
+  ensure_search_dialog (guw);
+  gucharmap_search_dialog_set_search (GUCHARMAP_SEARCH_DIALOG (guw->search_dialog), str);
+  gucharmap_search_dialog_start_search (GUCHARMAP_SEARCH_DIALOG (guw->search_dialog),
+                                        GUCHARMAP_DIRECTION_FORWARD);
+}
